-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
--Mandatory controller parameters
M.type = "auxiliary"
M.relevantDevice = nil
M.engineInfo = {}
M.gearboxInfo = {isReverse = false}
M.fireEngineTemperature = 0
M.throttle = 0
M.brake = 0
M.clutchRatio = 0

M.weaponController = nil

local min = math.min --pull math.min/max to local scope for faster access (and less text)
local max = math.max
local abs = math.abs

local engineNodes = {L = {}, R = {}}
local barrelNodes = {front = {}, back= {}}
local leftEngine = nil
local rightEngine = nil

local gearTimer = 0

local energyStorageData = {
  tanks = {},
  ratio = 0,
  volume = 0,
  capacity = 0,
  invEnergyStorageCount = 0,
}

local bulletBeams = {}
local bulletNodes = {}
local activeBullets = {}
local activeBulletCount = 0
--local recoilCounter = 0
local muzzleFlashCounter = 0
local rofTimer = 0
local numberOfRounds = nil
local roundIndex = nil
local fireRateTime = nil
local bulletTargetSpeed = 0
local realBulletSpeed = 540
local realBulletSpeedSquared = realBulletSpeed * realBulletSpeed
local quarterBulletTargetSpeed = 0
local additionalExplosiveEnergy = 0 --J

local roundsLeft = nil

local function updateBulletFiring(bullet, dt)
  local nodeVelocity = obj:getNodeVelocityVector(bullet.nodeCid):length() - bullet.velocityOffset
  if nodeVelocity >= bulletTargetSpeed then
    electrics.values[bullet.thrusterSource] = nil
    --recoilCounter = recoilCounter - 1
    local nodeMass = obj:getNodeMass(bullet.nodeCid)
    local targetEnergy = 0.5 * nodeMass * realBulletSpeedSquared + additionalExplosiveEnergy
    local equivalentWeight = min(max(targetEnergy * 2 / (nodeVelocity * nodeVelocity), nodeMass), 100)
    obj:setNodeMass(bullet.nodeCid, equivalentWeight)
    bullet.roundState = "fired"
    bullet.update = nop
  end
end

local function update(dt)
  for i = 1, activeBulletCount, 1 do
    activeBullets[i]:update(dt)
  end

  --electrics.values.recoil = recoilCounter > 0 and 1 or nil --activates the recoil thruster
end

local function toggleGuns()
	--M.screenData.gunState = electrics.values.gundoors
	electrics.values.gundoors = 1 - (electrics.values.gundoors or 0)
	if electrics.values.gundoors > 0 then 
		-- if M.screenData.mode then nextScreen(3) end
		if v.userPartConfig.moonhawk_6x6_F_display ~= "none" then
			controller.getController("old_clock").updateAmmo(roundsLeft)
			controller.getController("old_clock").nextScreen(3)
		end
		gui.message("Guns enabled", 3, "gunStatus")
	else
		-- if M.screenData.mode == 3 then nextScreen(1) end
		if v.userPartConfig.moonhawk_6x6_F_display ~= "none" then
			controller.getController("old_clock").nextScreen(1)
		end
		gui.message("Guns disabled", 3, "gunStatus")
	end
end

local function updateGFX(dt)
  --------GUNS--------
  --print(recoilCounter)
  local activeBullets2 = {}
  for _,v in ipairs(activeBullets) do
    v.muzzleFlashTimer = max(v.muzzleFlashTimer - dt, 0)
    if v.muzzleFlashTimer <= 0 and v.hasMuzzleFlash then
      muzzleFlashCounter = muzzleFlashCounter - 1
      v.hasMuzzleFlash = false
    end
    if v.roundState == "fired" then
      local bulletVelocity = abs(obj:getNodeVelocityVector(v.nodeCid):length())
      if bulletVelocity <= quarterBulletTargetSpeed then
        obj:breakBeam(v.beamCid or 0)
        --obj:addParticleByNodes(v.nodeCid, 0.1, 0, 31, 0, 1)
        if v.hasMuzzleFlash then
          muzzleFlashCounter = muzzleFlashCounter - 1
          v.hasMuzzleFlash = false
        end
        v.roundState = "dead"
      end
    end
    if v.roundState ~= "dead" then
      table.insert(activeBullets2, v)
    end
  end

  activeBullets = activeBullets2
  --dump(activeBullets)
  
  -- if (M.screenData.gunState ~= electrics.values.gundoors) then toggleGuns() end

  rofTimer = max(rofTimer - dt, 0)
  if electrics.values.guns >= 1 and electrics.values.gundoors >= 0.8 and rofTimer <= 0 and roundIndex < numberOfRounds then --only go in here if we are STILL firing or want to fire
	sounds.playSoundOnceAtNode("moonhawk_6x6_machinegun", barrelNodes.front[1], 5)	--Play the firing sound
	--sounds.playSoundOnceAtNode("moonhawk_6x6_machinegun", barrelNodes.front[2], 5)
	--sounds.playSoundOnceAtNode("moonhawk_6x6_machinegun", barrelNodes.front[3], 5)
	sounds.playSoundOnceAtNode("moonhawk_6x6_machinegun", barrelNodes.front[4], 5)

	--obj:addParticleByNodesRelative(barrelNodes.front[1], barrelNodes.back[1], 10, 6, 0, 1)
	--obj:addParticleByNodesRelative(barrelNodes.front[2], barrelNodes.back[2], 10, 6, 0, 1)
	--obj:addParticleByNodesRelative(barrelNodes.front[3], barrelNodes.back[3], 10, 6, 0, 1)
	--obj:addParticleByNodesRelative(barrelNodes.front[4], barrelNodes.back[4], 10, 6, 0, 1)
	
	roundsLeft = (numberOfRounds - roundIndex - 1) * 4
	if v.userPartConfig.moonhawk_6x6_F_display ~= "none" then
		controller.getController("old_clock").updateAmmo(roundsLeft)
	end
	
	gui.message(string.format("Ammunition: %d", roundsLeft), 3, "cannonrounds")
	for i = 1, 4, 1 do
	  local thrusterSource = "shell_"..(roundIndex * 4 + i)
	  electrics.values[thrusterSource] = 1 --roundIndex starts at 1, and determines which bullet is to be fired. it is added to "shell_", so if roundIndex is 1 then the thruster with "bllt_1" as the control is activated
	  table.insert(activeBullets,
		{
		  beamCid = bulletBeams[roundIndex * 4 + i],
		  nodeCid = bulletNodes[roundIndex * 4 + i],
		  thrusterSource = thrusterSource,
		  velocityOffset = electrics.values.airspeed,
		  muzzleFlashTimer = 0.05,
		  hasMuzzleFlash = true,
		  roundState = "firing",
		  update = updateBulletFiring})
	end
	rofTimer = fireRateTime
	roundIndex = roundIndex + 1
	--recoilCounter = recoilCounter + 4
	muzzleFlashCounter = muzzleFlashCounter + 4
	
  end

  activeBulletCount = #activeBullets
  --print(activeBulletCount)
  --dump(activeBullets)
  electrics.values.muzf = muzzleFlashCounter > 0 and 1 or 0	--show the muzzleflash. This is done with glowmaps with "muzf" as their input.
  electrics.values.recoil = electrics.values.muzf
end

local function init(jbeamData)
  electrics.values.guns = 0 --gun trigger
  electrics.values.recoil = nil --gun recoil
  electrics.values.muzf = 0 --muzzle flash
  electrics.values.gundoors = 0 --doors
  --recoilCounter = 0
  muzzleFlashCounter = 0
  roundIndex = 0
  numberOfRounds = jbeamData.numberOfRounds or 0
  roundsLeft = numberOfRounds * 4
  fireRateTime = 0.09
  bulletTargetSpeed = jbeamData.bulletSpeed or 410
  quarterBulletTargetSpeed = bulletTargetSpeed * 0.25
  additionalExplosiveEnergy = jbeamData.bulletAdditionalExplosiveEnergy or 536000
  
  barrelNodes = {front = {}, back= {}}

  for k, node in pairs (v.data.nodes) do
    if node.name == "barrel_1_f" then barrelNodes.front[1] = k end
    if node.name == "barrel_2_f" then barrelNodes.front[2] = k end
    if node.name == "barrel_3_f" then barrelNodes.front[3] = k end
    if node.name == "barrel_4_f" then barrelNodes.front[4] = k end
    if node.name == "barrel_1_b" then barrelNodes.back[1] = k end
    if node.name == "barrel_2_b" then barrelNodes.back[2] = k end
    if node.name == "barrel_3_b" then barrelNodes.back[3] = k end
    if node.name == "barrel_4_b" then barrelNodes.back[4] = k end
    if node.bullet then bulletNodes[node.bullet] = node.cid  end
  end

  activeBullets = {}
  activeBulletCount = 0
  bulletBeams = {}
  
  for _, beam in pairs(v.data.beams) do
    if beam.tracer then bulletBeams[beam.tracer] = beam.cid  end
  end
end

M.init = init
M.initSecondStage = initSecondStage
M.updateGFX = updateGFX
M.update = update
M.nextScreen = nextScreen
M.toggleGuns = toggleGuns

--Mandatory main controller API
M.shiftUp = nop
M.shiftDown = nop
M.shiftToGearIndex = nop
M.cycleGearboxModes = nop
M.setGearboxMode = nop
M.setStarter = nop
M.setFreeze = nop
M.sendTorqueData = nop
M.vehicleActivated = nop
-------------------------------

return M